# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import itertools
import logging
import os.path
import re
import sys

import expatriate
from publishsubscribe import Subscriber

from .decorators import *
from .exceptions import *

logger = logging.getLogger(__name__)
#logger.setLevel(logging.INFO)

@attribute(namespace='http://www.w3.org/XML/1998/namespace', local_name='lang', type=('expatriate.model.xs.StringType', 'StringType'), into='_xml_lang')
@attribute(namespace='http://www.w3.org/XML/1998/namespace', local_name='space', enum=('default', 'preserve'), into='_xml_space')
@attribute(namespace='http://www.w3.org/XML/1998/namespace', local_name='base', type=('expatriate.model.xs.AnyUriType', 'AnyUriType'), into='_xml_base')
@attribute(namespace='http://www.w3.org/XML/1998/namespace', local_name='id', type=('expatriate.model.xs.IdType', 'IdType'), into='_xml_id')
@attribute(namespace='http://www.w3.org/2000/xmlns/', local_name='*', into='_xmlns')
@attribute(namespace='http://www.w3.org/2001/XMLSchema-instance', local_name='type', type=('expatriate.model.xs.QNameType', 'QNameType'), into='_xsi_type')
@attribute(namespace='http://www.w3.org/2001/XMLSchema-instance', local_name='nil', type=('expatriate.model.xs.BooleanType', 'BooleanType'), into='_xsi_nil', default=False)
@attribute(namespace='http://www.w3.org/2001/XMLSchema-instance', local_name='schemaLocation', type=('expatriate.model.xs.AnyUriType', 'AnyUriType'), into='_xsi_schemaLocation')
@attribute(namespace='http://www.w3.org/2001/XMLSchema-instance', local_name='noNamespaceSchemaLocation', type=('expatriate.model.xs.AnyUriType', 'AnyUriType'), into='_xsi_noNamespaceSchemaLocation')
@content()
class Model(Subscriber):
    '''
    Base class for a XML object relation model. Allows mapping XML data into Python types.

    :param local_name: Override for the xml local name when Model is generated.
    :type local_name: str or None
    :param namespace: Override for the xml namespace when Model is generated.
    :type namespace: str or None
    :param prefix: Override for the xml prefix when Model is generated.
    :type prefix: str or None
    :param value: Value to add to the Model's children (generated as CharacterData). No child is added if left as the default None.
    :type local_name: str or None
    :raises UnknownNamespaceException: if unable to determine a namespace model maps to
    '''

    ANY_NAMESPACE = '*'
    ''' Identifier representing any namespace '''

    ANY_LOCAL_NAME = '*'
    ''' Identifier representing any local name '''

    __namespace_to_package = {
        'http://www.w3.org/XML/1998/namespace': 'expatriate.model.xml',
        'http://www.w3.org/2001/XMLSchema': 'expatriate.model.xs',
        'http://www.w3.org/2001/XMLSchema-hasFacetAndProperty': 'expatriate.model.xshfp',
        'http://www.w3.org/2001/XMLSchema-instance': 'expatriate.model.xsi',
    }
    __namespace_to_prefix = {
        'http://www.w3.org/XML/1998/namespace': 'xml',
        'http://www.w3.org/2001/XMLSchema': 'xs',
        'http://www.w3.org/2001/XMLSchema-hasFacetAndProperty': 'xshfp',
        'http://www.w3.org/2001/XMLSchema-instance': 'xsi',
    }
    __package_to_namespace = {v:k for k,v in __namespace_to_package.items()}
    _ns_count = 0

    _attribute_mappers = {}
    _attribute_mapper_cache = {}
    _element_mappers = {}
    _element_mapper_cache = {}
    _element_mapper_order = {}
    _content_mappers = {}
    _content_mapper_cache = {}

    def __init__(self, local_name=None, namespace=None, prefix=None, value=None):
        self._parent = None
        self._children = []

        if value is not None:
            self._children.append((None, value))

        self._local_name = local_name

        if namespace is not None:
            self._namespace = namespace
        else:
            self._namespace = Model.package_to_namespace(self.get_package())
            if self._namespace is None:
                # make sure namespace is never None
                raise UnknownNamespaceException('Unable to determine namespace for xml generation: ' + str(self))

        if prefix is not None:
            self._prefix = prefix
        else:
            self._prefix = Model.namespace_to_prefix(self._namespace)

        at_mappers = self._get_attribute_mappers()
        el_mappers = self._get_element_mappers()
        content_mappers = self._get_content_mappers()
        for mapper in itertools.chain(at_mappers, el_mappers, content_mappers):
            mapper.initialize(self)
        self._initialized = True

    @staticmethod
    def class_for_element(el):
        '''
        load the model class corresponding to an element

        :param expatriate.Element el: The element for which to look up the class.
        :raises ElementMappingException: if class for element cannot be discovered
        '''

        model_package = Model.namespace_to_package(el.namespace)
        try:
            pkg_mod = importlib.import_module(model_package)
        except:
            raise ElementMappingException('Unable to determine mapping for '
                + str(el) + ' element: cannot load package module '
                + model_package)

        try:
            class_name = pkg_mod.ELEMENT_MAP[el.namespace, el.local_name]
        except AttributeError:
            raise ElementMappingException(pkg_mod.__name__
                + ' does not define ELEMENT_MAP; cannot load ' + str(el))
        except KeyError:
            raise ElementMappingException(pkg_mod.__name__
                + ' does not define mapping for ' + str(el) + ' element')

        try:
            mod = importlib.import_module(model_package + '.' + class_name)
            class_ = getattr(mod, class_name)
        except:
            raise ElementMappingException('Unable to determine mapping for '
                + str(el) + ' element: cannot load class module '
                + model_package + '.' + class_name)

        return class_

    @classmethod
    def _get_attribute_mappers(cls):
        '''
        get all the attribute definitions for a model class
        '''

        if cls.__name__ not in Model._attribute_mapper_cache:
            mappers = []

            for cls_ in reversed(cls.__mro__):
                if issubclass(cls_, Model):
                    try:
                        logger.debug('Adding attribute mappers from superclass '
                            + cls_.__name__)
                        mappers.extend(cls_._attribute_mappers[cls_.__name__])
                    except KeyError:
                        pass
            Model._attribute_mapper_cache[cls.__name__] = mappers

        return Model._attribute_mapper_cache[cls.__name__]

    @classmethod
    def _add_attribute_mapper(cls, mapper):
        '''
        set the model attribute definition for an attribute
        '''

        if cls.__name__ not in cls._attribute_mappers:
            cls._attribute_mappers[cls.__name__] = []

        # mappers come in reverse order, so we insert rather than append
        cls._attribute_mappers[cls.__name__].insert(0, mapper)

    @classmethod
    def _get_element_mappers(cls):
        '''
        get all the element definitions for a model class
        '''

        if cls.__name__ not in Model._element_mapper_cache:
            mappers = []

            for cls_ in reversed(cls.__mro__):
                if issubclass(cls_, Model):
                    try:
                        logger.debug('Adding element mappers from superclass '
                            + cls_.__name__)
                        mappers.extend(cls_._element_mappers[cls_.__name__])
                    except KeyError:
                        pass

            Model._element_mapper_cache[cls.__name__] = mappers

        return Model._element_mapper_cache[cls.__name__]

    @classmethod
    def _add_element_mapper(cls, mapper):
        '''
        set the model element definition for an element
        '''

        if cls.__name__ not in cls._element_mappers:
            cls._element_mappers[cls.__name__] = []

        # mappers come in reverse order, so we insert rather than append
        cls._element_mappers[cls.__name__].insert(0, mapper)

        # now set the order that this element was defined
        if cls.__name__ not in cls._element_mapper_order:
            cls._element_mapper_order[cls.__name__] = []

        # have to insert at the front because decorators are applied in reverse
        # order
        cls._element_mapper_order[cls.__name__].insert(
            0,
            (mapper.get_namespace(), mapper.get_local_name())
        )

    @classmethod
    def _add_content_mapper(cls, mapper):
        '''
        add a model content definition for the class
        '''

        if cls.__name__ not in cls._content_mappers:
            cls._content_mappers[cls.__name__] = []

        # mappers come in reverse order, so we insert rather than append
        cls._content_mappers[cls.__name__].insert(0, mapper)

    @classmethod
    def _get_content_mappers(cls):
        '''
        get the model content definitions
        '''

        if cls.__name__ not in Model._content_mapper_cache:
            mappers = []

            for cls_ in reversed(cls.__mro__):
                if issubclass(cls_, Model):
                    try:
                        logger.debug('Adding content mappers from superclass '
                            + cls_.__name__)
                        mappers.extend(cls_._content_mappers[cls_.__name__])
                    except KeyError:
                        pass

            Model._content_mapper_cache[cls.__name__] = mappers

        return Model._content_mapper_cache[cls.__name__]

    @staticmethod
    def register_namespace(model_package, namespace, prefix=None):
        '''
        Register a namespace for use

        :param str model_package: The package to use for mapping elements from namespace
        :param str namespace: The namespace
        :param prefix: The prefix to use when mapping to namespace
        :type prefix: str or None
        '''
        Model.__namespace_to_package[namespace] = model_package
        Model.__package_to_namespace[model_package] = namespace
        Model.__namespace_to_prefix[namespace] = prefix

    @staticmethod
    def unregister_namespace(model_package):
        '''
        Unregister a namespace.

        :param str model_package: The package to unregister from mapping
        :raises UnknownNamespaceException: if the namspace isn't registered
        '''
        try:
            namespace = Model.__package_to_namespace[model_package]
        except KeyError:
            raise UnknownNamespaceException('Unregistered namespace: '
                + model_package)

        del Model.__package_to_namespace[model_package]
        del Model.__namespace_to_package[namespace]
        del Model.__namespace_to_prefix[namespace]

    @staticmethod
    def package_to_namespace(model_package):
        '''
        Find namespace corresponding to a model package.

        :param str model_package: The package to lookup
        :raises UnknownNamespaceException: if the package isn't registered
        '''
        logger.debug('Looking for xml namespace for model package '
            + model_package)
        if model_package not in Model.__package_to_namespace:
            raise UnknownNamespaceException('Package ' + model_package
                + ' is not in registered packages')

        return Model.__package_to_namespace[model_package]

    @staticmethod
    def namespace_to_package(namespace):
        '''
        Find model package corresponding to namespace.

        :param str namespace: The namespace to lookup
        :raises UnknownNamespaceException: if the namspace isn't registered
        '''
        logger.debug('Looking for model package for xml namespace ' + str(namespace))
        if namespace not in Model.__namespace_to_package:
            raise UnknownNamespaceException('XML namespace ' + str(namespace)
                + ' is not in registered namespaces')

        return Model.__namespace_to_package[namespace]

    @staticmethod
    def namespace_to_prefix(namespace):
        '''
        Find the model package corresponding to a namespace.

        :param str namespace: The namespace to lookup
        :raises UnknownNamespaceException: if the namspace isn't registered
        '''
        logger.debug('Looking for xml prefix for xml namespace ' + str(namespace))

        prefix = None

        try:
            prefix = Model.__namespace_to_prefix[namespace]
        except KeyError:
            prefix = 'ns' + str(Model._ns_count)
            Model._ns_count += 1

            logger.info(pkg_mod.__name__
                + ' did not register prefix; generated: ' + prefix)
        except:
            raise UnknownNamespaceException('Unable to determine prefix for '
                + namespace + ' namespace')

        return prefix

    @classmethod
    def get_package(cls):
        '''
        Find package corresponding to class/instance

        :rtype: str
        '''
        return cls.__module__.rpartition('.')[0]

    @staticmethod
    def load(parent, el):
        '''
        load a Model given an expatriate Element

        :param expatriate.model.Model parent: The :py:class:`.Model` to use as the parent of the model we're mapping from el
        :param expatriate.Element el: The :py:class:`..Element` from which we're mapping
        '''

        # try to load the element's module
        if parent is None:
            if el.namespace is None:
                raise UnknownNamespaceException(
                    'Unable to determine namespace without fully qualified element ('
                    + str(el) + ') and parent model')

            class_ = Model.class_for_element(el)
        else:
            logger.debug('Checking ' + parent.__class__.__name__
                + ' for element ' + str(el))

            for mapper in parent._get_element_mappers():
                if mapper.matches(el, parent):
                    logger.debug(str(el) + ' matched ' + str(mapper)
                        + ' in ' + parent.__class__.__name__)
                    class_ = mapper.class_for_element(el, parent)
                    break
            else:
                raise ElementMappingException(parent.__class__.__name__
                    + ' does not define mapping for '
                    + str(el) + ' element')

        logger.debug('Loaded class ' + str(class_) + ' for ' + str(el))

        # instantiate an instance of the class & load it
        inst = class_()
        inst.parse(parent, el)

        return inst

    @staticmethod
    def find_content(uri):
        '''
        Loads the content from the URI, parses it, returns the root Model.
        Remote URIs are not yet supported.

        :param str uri: The URI to load
        :rtype: .Model or subclass
        :raises ReferenceException: if the content could not be found
        '''

        if os.path.isfile(uri):
            try:
                doc = expatriate.Document()
                doc.parse_file(uri)
                return Model.load(None, doc.root_element)
            except:
                raise ReferenceException('Could not find content for: ' + uri)
        else:
            raise NotImplementedError('Remote URI loading is not yet implemented')

        raise ReferenceException('Could not find content for: ' + uri)

    def __setattr__(self, name, value):
        '''
        setattr override to keep track of indexes etc.
        '''

        if not hasattr(self, '_initialized'):
            object.__setattr__(self, name, value)
            return

        el_mappers = self._get_element_mappers()
        for mapper in itertools.chain(el_mappers):
            if mapper.get_attr_name() == name:
                mapper.setattr(self, name, value)
                return

        object.__setattr__(self, name, value)

    def _attr_name_from_publisher(self, publisher):
        for k, v in self.__dict__.items():
            if publisher == v:
                return k
        raise AttributeError('Attribute matching ' + str(publisher) + ' not found')

    def data_added(self, publisher, id_, item):
        '''
        Receive notification from a Publisher when data has been added

        :param publishsubscribe.Publisher publisher: The publisher
        :param id_: The implementation dependent id of the data added; list index or dict key for example
        :param item: The item added
        '''
        pub_name = self._attr_name_from_publisher(publisher)
        self._children.append((pub_name, id_))

    def data_updated(self, publisher, id_, old_item, new_item):
        '''
        Receive notification from a Publisher when data has been updated

        :param publishsubscribe.Publisher publisher: The publisher
        :param id_: The implementation dependent id of the data updated; list index or dict key for example
        :param old_item: The old item
        :param new_item: The new item
        '''
        pass

    def data_deleted(self, publisher, id_, item):
        '''
        Receive notification from a Publisher when data has been deleted

        :param publishsubscribe.Publisher publisher: The publisher
        :param id_: The implementation dependent id of the data deleted; list index or dict key for example
        :param item: The item deleted
        '''
        pub_name = self._attr_name_from_publisher(publisher)
        self._children.remove((pub_name, id_))
        if isinstance(publisher, list):
            # need to update the list indexes
            new_children = []
            for attr_name, idx in self._children:
                if pub_name == attr_name and isinstance(idx, int) and idx > id_:
                    new_children.append((attr_name, idx -1))
                else:
                    new_children.append((attr_name, idx))
            object.__setattr__(self, '_children', new_children)

    def is_nil(self):
        '''
        Determine if the model's xsi nil is set

        :rtype: bool
        '''
        return self._xsi_nil

    def set_nil(self):
        '''
        Set model's xsi nil and clear the value.
        '''
        self._xsi_nil = True
        self.set_value(None)

    def _get_content_children(self):
        '''
        Get the content children of the model
        '''
        content = []
        for attr_name, id_ in self._children:
            if attr_name is None:
                content.append(id_)
        return content

    def get_value(self):
        '''
        Get the joined str value of the model's children

        :rtype: str
        '''
        content = self._get_content_children()
        if len(content) == 0:
            return None
        else:
            return ''.join(content)

    def set_value(self, value):
        '''
        Set content to *value*. Effectively clears all content children and
        appends the *value*.

        :param str value: The str value
        '''
        new_children = []
        for attr_name, id_ in self._children:
            if attr_name is not None:
                new_children.append((attr_name, id_))
        self._children = new_children
        if value is not None:
            self._children.append((None, value))

    def parse_value(self, value):
        '''
        Parse the given *value* and return; overriden for value-limiting
        subclasses (most of the :py:mod:`expatriate.model.xs` \*Type classes)

        For example :py:class:`expatriate.model.xs.FloatType` parses a
        :py:class:`str` value and returns a :py:class:`float` type

        :param value: The value to be parsed
        '''
        return value

    def produce_value(self, value):
        '''
        produce the given *value* and return; overriden for value-limiting
        subclasses (most of the :py:mod:`expatriate.model.xs` \*Type classes)

        For example :py:class:`expatriate.model.xs.FloatType` produces a
        :py:class:`str` type from a :py:class:`float` value

        :param value: The value produced
        '''
        if value is None:
            return value
        return str(value)

    def __str__(self):
        s = self.__class__.__name__
        if hasattr(self, 'id') and self.id is not None:
            s += ' id: ' + self.id
        elif hasattr(self, 'Id') and self.Id is not None:
            s += ' Id: ' + self.Id
        elif hasattr(self, 'name') and self.name is not None:
            s += ' name: ' + self.name
        else:
            s += ' # ' + str(id(self))

        return s

    def find_reference(self, ref):
        '''
        Find child that matches reference *ref*. A child matches if it has an
        attribute *id* that is equal to the ref argument.

        :param str ref: The reference to a child for which to search this model recursively
        :rtype: .Model
        '''

        logger.debug('Matching reference ' + ref + ' against ' + str(self))

        try:
            if self.id == ref:
                return self
        except AttributeError:
            pass

        # check element mappers for ref
        for mapper in self._get_element_mappers():
            try:
                return mapper.find_reference_in(ref, self)
            except ReferenceException:
                pass

        raise ReferenceException('Could not find reference ' + ref
            + ' within ' + str(self))

    def parse(self, parent, el):
        '''
        Load model data from xml element *el*

        :param .Model parent: The parent of the model being parsed
        :param expatriate.Element el: The element being parsed into the model
        '''

        self._parent = parent

        self._local_name = el.local_name
        self._namespace = el.namespace
        self._prefix = el.prefix

        logger.debug('Parsing ' + str(el) + ' element into '
            + self.__class__.__module__ + '.' + self.__class__.__name__
            + ' class')

        for name, attr in el.attribute_nodes.items():
            for mapper in self._get_attribute_mappers():
                if mapper.matches(attr, self):
                    mapper.parse_in(self, attr)
                    break
            else:
                # if we didn't find a match for the attribute, raise
                raise UnknownAttributeException('Unknown ' + str(self)
                + ' attribute ' + str(attr))

        for child in el.children:
            if isinstance(child, expatriate.Element):
                for mapper in self._get_element_mappers():
                    if mapper.matches(child, self):
                        mapper.parse_in(self, child)
                        break
                else:
                    raise UnknownElementException('Unknown ' + str(self)
                        + ' child ' + str(child) + ' does not match any mappers')
            else:
                self._children.append((None, child.get_string_value()))

        at_mappers = self._get_attribute_mappers()
        el_mappers = self._get_element_mappers()
        content_mappers = self._get_content_mappers()
        for mapper in itertools.chain(at_mappers, el_mappers, content_mappers):
            mapper.validate(self)

    def produce(self, parent_el=None):
        '''
        Generate xml from the model

        :param expatriate.Element parent_el: The expatriate.Element to use as the parent of the element generated from this model
        :rtype: expatriate.Element
        '''

        logger.debug(str(self) + ' to xml')

        at_mappers = self._get_attribute_mappers()
        el_mappers = self._get_element_mappers()
        content_mappers = self._get_content_mappers()
        for mapper in itertools.chain(at_mappers, el_mappers, content_mappers):
            mapper.validate(self)

        if self._local_name is None:
            raise ElementMappingException('local_name must be defined by constructor or @element')

        el = expatriate.Element(self._local_name, namespace=self._namespace, prefix=self._prefix, parent=parent_el)

        for mapper in at_mappers:
            mapper.produce_in(el, self)

        for attr_name, id_ in self._children:
            if attr_name is None:
                for mapper in content_mappers:
                    mapper.produce_in(el, self, id_)
            else:
                for mapper in el_mappers:
                    if attr_name == mapper.get_attr_name():
                        mapper.produce_in(el, self, id_)
                        break # for mapper...
                else: # for mapper
                    raise ElementMappingException('Unable to map ' + str(self)
                        + ' attribute ' + attr_name + str([id_]) + ' to element')

        return el



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import logging
import math
import re

from .exceptions import *
from .xpath import *

logger = logging.getLogger(__name__)

class Node(object):
    '''
    Class representing a XML node

    :keyword parent: parent Node of this Node
    :type parent: Parent or None

    '''
    def __init__(self, parent=None):
        self._parent = parent

    @property
    def parent(self):
        return self._parent

    @parent.setter
    def parent(self, parent):
        self._parent = parent

    def prefix_to_namespace(self, prefix):
        if prefix == 'xmlns':
            return 'http://www.w3.org/2000/xmlns/'
        elif prefix == 'xml':
            return 'http://www.w3.org/XML/1998/namespace'
        elif isinstance(self._parent, Node):
            return self._parent.prefix_to_namespace(prefix)
        else:
            raise UnknownPrefixException('Unknown prefix: ' + str(prefix))

    def namespace_to_prefix(self, namespace_uri):
        if namespace_uri == 'http://www.w3.org/2000/xmlns/':
            return 'xmlns'
        elif namespace_uri == 'http://www.w3.org/XML/1998/namespace':
            return 'xml'
        elif isinstance(self._parent, Node):
            return self._parent.namespace_to_prefix(namespace_uri)
        else:
            raise UnknownNamespaceException('Unknown namespace uri: ' + str(namespace_uri))

    def _tokenize(self, expr):
        tokens = []

        parens = 0
        braces = 0

        t = ''
        for i, char in enumerate(expr):
            if len(t) > 0:
                if t[0] in '\'"':
                    # string literal
                    t += char
                    if char == t[0] and t[-1] != '\\':
                        tokens.append(t)
                        t = ''
                elif t == '-':
                    tokens.append(t)
                    t = char
                elif t == '.' and char != '.':
                    tokens.extend(['self', '::', 'node', '(', ')'])
                    t = char
                elif re.fullmatch(r'[0-9][0-9.]*', t) and (char.isdigit() or char == '.'):
                    t += char
                elif t[0] == '$' and char.isalpha():
                    t += char
                elif t.isspace():
                    # skip space
                    t = char
                elif t in ':/.!<>':
                    if t + char in ['::', '//', '..', '!=', '<=', '>=']:
                        t += char
                        if t == '//':
                            tokens.extend(['/', 'descendant-or-self', '::', 'node', '(', ')', '/'])
                        elif t == '..':
                            tokens.extend(['parent', '::', 'node', '(', ')'])
                        else:
                            tokens.append(t)
                        t = ''
                    else:
                        tokens.append(t)
                        t = char
                elif t == '@':
                    tokens.extend(['attribute', '::'])
                    t = char
                elif t == '(':
                    tokens.append(t)
                    t = char
                    parens += 1
                elif t == ')':
                    tokens.append(t)
                    t = char
                    parens -= 1
                elif t == '[':
                    tokens.append(t)
                    t = char
                    braces += 1
                elif t == ']':
                    tokens.append(t)
                    t = char
                    braces -= 1
                elif t in ',\'"*|+=':
                    tokens.append(t)
                    t = char
                elif char.isalnum() or char == '-':
                    t += char
                else:
                    tokens.append(t)
                    t = char
            else:
                if char.isspace():
                    continue
                t += char

        # append final token if there is one
        if t == '.':
            tokens.extend(['self', '::', 'node', '(', ')'])
        elif t.isspace():
            pass
        elif t == ')':
            tokens.append(t)
            parens -= 1
        elif t == ']':
            tokens.append(t)
            braces -= 1
        elif t != '':
            tokens.append(t)

        if parens != 0 or braces != 0:
            raise XPathSyntaxException('Paren or brace expression not closed')

        return tokens

    def xpath(self, expr, version=1.0, variables={}, add_functions={}):
        if version != 1.0:
            raise NotImplementedError('Only XPath 1.0 has been implemented')

        if self.get_document() is None:
            raise ValueError("Can't resolve xpath expression on Node not attached to a document")

        logger.debug('********************************************')
        logger.debug('Tokenizing xpath expression: ' + str(expr))

        tokens = self._tokenize(expr)
        logger.debug('Tokens: ' + str(tokens))

        functions = Function.FUNCTIONS.copy()
        functions.update(add_functions)

        stack = []
        for i, token in enumerate(tokens):
            if token == '(':
                e = Expression()
                logger.debug('Starting sub expression ' + str(e))
                stack.append(e)
            elif token == ')':
                if len(stack) <= 1:
                    continue

                logger.debug('End of ' + str(stack[-1]))
                if i > 0 and tokens[i-1] == '(':
                    # don't add empty Expression
                    stack.pop()
                    logger.debug('Ignoring empty expression')
                elif len(stack) > 1:
                    e = stack.pop()
                    logger.debug('Adding ' + str(e) + ' to ' + str(stack[-1]))
                    stack[-1].children.append(e)
                # else just let it on the stack
            elif token == '[':
                p = Predicate()
                stack.append(p)
                logger.debug('Starting predicate ' + str(p))
                e = Expression()
                stack.append(e)
                logger.debug('Starting sub expression ' + str(e))
            elif token == ']':
                logger.debug('End of ' + str(stack[-1]))
                if i > 0 and tokens[i-1] == '[':
                    # don't add empty predicate
                    stack.pop()
                    stack.pop()
                    logger.debug('Ignoring empty expression & predicate')
                else:
                    while(len(stack) > 1 and not isinstance(stack[-1], Predicate)):
                        i = stack.pop()
                        logger.debug('Adding ' + str(i) + ' to ' + str(stack[-1]))
                        stack[-1].children.append(i)
                    p = stack.pop()
                    if not isinstance(stack[-1], Axis):
                        raise SyntaxException('Expecting Axis on stack before predicate')
                    logger.debug('Adding ' + str(p) + ' to ' + str(stack[-1]))
                    stack[-1].children.append(p)
            elif token  == '::':
                # already processed axis
                pass
            elif token == ':':
                # already processed QNameNodeTest
                pass
            elif token == '*':
                if i > 0 and tokens[i-1] not in [
                    '::', '(', '[', ',', 'and', 'or', 'mod', 'div',
                    '*', '/', '//', '|', '+', '-', '=', '!=', '<', '<=',
                    '>', '>=']:
                    o = Operator(token)
                    o.children.append(stack.pop())
                    stack.append(o)
                    logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
                else:
                    if len(stack) == 0 or not isinstance(stack[-1], Axis):
                        # use implicit axis
                        a = Axis('child')
                        stack.append(a)
                        logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
                    nt = AnyNodeTest(stack[-1].get_principal_node_type())
                    stack[-1].children.append(nt)
                    logger.debug('Adding ' + str(nt) + ' to children of ' + str(stack[-1]))
            elif token == ',':
                try:
                    while(not isinstance(stack[-1], Function)):
                        i = stack.pop()
                        logger.debug('Adding ' + str(i) + ' to children of ' + str(stack[-1]))
                        stack[-1].children.append(i)
                except IndexError:
                    raise SyntaxException('Unable to add argument to function')

                logger.debug('Starting new expression ' + str(e) + ' for function argument')
                e = Expression()
                stack.append(e)
                logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
            elif token[0] in '\'"':
                l = Literal(token[1:-1])
                stack.append(l)
                logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
            elif re.fullmatch(r'[0-9.]+', token):
                if '.' in token:
                    l = Literal(float(token))
                else:
                    l = Literal(int(token))

                if len(stack) > 0 and isinstance(stack[-1], Operator):
                    op = stack.pop()
                    op.children.append(l)
                    logger.debug('Added ' + str(l) + ' to children of ' + str(op))
                    stack.append(op)
                else:
                    stack.append(l)
                logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
            elif token == '-' and ( \
                len(stack) == 0 \
                or isinstance(stack[-1], Operator) \
                or (i > 0 and tokens[i-1] in ('(', ',', '[')) \
            ):
                o = Operator('negate')
                stack.append(o)
                logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
            elif token in Operator.OPERATORS and i > 0 and tokens[-1] not in [
                '::', '(', '[', ',', 'and', 'or', 'mod', 'div',
                '*', '/', '//', '|', '+', '-', '=', '!=', '<', '<=',
                '>', '>=',
            ]:
                o = Operator(token)
                o.children.append(stack.pop())
                stack.append(o)
                logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
            elif token == '/':
                if i == 0:
                    s = RootStep(self.get_document())
                else:
                    while(len(stack) > 0 and not isinstance(stack[-1], Step)):
                        i = stack.pop()
                        if len(stack) > 0:
                            logger.debug('Adding ' + str(i) + ' to children of ' + str(stack[-1]))
                            stack[-1].children.append(i)

                    if len(stack) == 0 or not isinstance(stack[-1], Step):
                        logger.debug('Step is not the last item on the stack')
                        parent_step = Step()
                        logger.debug('Adding ' + str(i) + ' to children of ' + str(parent_step))
                        parent_step.children.append(i)
                        stack.append(parent_step)
                        logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
                    s = Step()
                stack.append(s)
                logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
            elif token == 'Infinity':
                stack.append(Literal(math.inf))
                logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
            elif token == 'NaN':
                stack.append(Literal(math.nan))
                logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
            elif re.fullmatch(r'[a-zA-Z0-9_-]+', token):
                if len(stack) > 0 and isinstance(stack[-1], Axis):
                    if token in TypeNodeTest.NODE_TYPES:
                        stack[-1].children.append(TypeNodeTest(token))
                        logger.debug('Added ' + str(stack[-1].children[-1]) + ' to children of ' + str(stack[-1]))
                    elif len(tokens) > i+1 and tokens[i+1] == ':':
                        stack.append(QNameNodeTest(token))
                        logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
                    else:
                        stack[-1].children.append(NCNameNodeTest(token))
                        logger.debug('Added ' + str(stack[-1].children[-1]) + ' to children of ' + str(stack[-1]))
                elif len(tokens) > i+1 and tokens[i+1] == '::':
                    if token not in Axis.AXES:
                        raise XPathSyntaxException('Unknown axis: ' + str(token))
                    a = Axis(token)
                    stack.append(a)
                    logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
                elif len(tokens) > i+1 and tokens[i+1] == '(':
                    if token in Function.FUNCTIONS:
                        f = Function(token, Function.FUNCTIONS[token])
                        stack.append(f)
                        logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
                    elif token in TypeNodeTest.NODE_TYPES:
                        if len(stack) == 0 or not isinstance(stack[-1], Axis):
                            stack.append(Axis('child'))
                            logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
                        nt = TypeNodeTest(token)
                        stack[-1].children.append(nt)
                        logger.debug('Added ' + str(nt) + ' to children of ' + str(stack[-1]))
                    else:
                        raise XPathSyntaxException('Unknown function or node type test: ' + str(token))
                elif token == 'true':
                    stack.append(Literal(True))
                    logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
                elif token == 'false':
                    stack.append(Literal(False))
                    logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
                elif token in Operator.OPERATORS and i > 0 and tokens[-1] not in [
                    '::', '(', '[', ',', 'and', 'or', 'mod', 'div', '*', '/',
                    '//', '|', '+', '-', '=', '!=', '<', '<=', '>', '>=']:
                    o = Operator(token)
                    o.children.append(stack.pop())
                    stack.append(o)
                    logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
                elif len(tokens) > i+1 and tokens[i+1] == ':':
                    # first part of a qname test
                    if len(stack) == 0 or not isinstance(stack[-1], Axis):
                        stack.append(Axis('child'))
                        logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
                    nt = QNameNodeTest(token)
                    stack.append(nt)
                    logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
                elif i > 0 and tokens[i-1] == ':':
                    # second part of a qname test
                    if len(stack) == 0 or not isinstance(stack[-1], QNameNodeTest):
                        raise SyntaxException('Expecting QNameNodeTest on stack; got second half of QName')
                    nt = stack.pop()
                    nt.name += ':' + token
                    if len(stack) == 0 or not isinstance(stack[-1], Axis):
                        raise SyntaxException('Expecting Axis on stack; finished QNameNodeTest')
                    stack[-1].children.append(nt)
                    logger.debug('Added ' + str(nt) + ' to ' + str(stack[-1]))
                else:
                    # has to be a ncname test
                    if len(stack) == 0 or not isinstance(stack[-1], Axis):
                        stack.append(Axis('child'))
                        logger.debug('Pushed ' + str(stack[-1]) + ' on stack')
                    nt = NCNameNodeTest(token)
                    stack[-1].children.append(nt)
                    logger.debug('Added ' + str(nt) + ' to children of ' + str(stack[-1]))
            else:
                raise XPathSyntaxException('Unknown token: ' + str(token))

        while(len(stack) > 1):
            i = stack.pop()
            logger.debug('Adding ' + str(i) + ' to children of ' + str(stack[-1]))
            stack[-1].children.append(i)
        i = stack.pop()
        logger.debug('Final pop off stack got ' + str(i))

        # TODO need to make sure node set items are unique
        logger.debug('********************************************')
        logger.debug('Evaluating ' + str(i))
        return i.evaluate(self, 1, 1, variables)

    def __str__(self):
        return self.__class__.__name__ + ' ' + hex(id(self))

    def __repr__(self):
        # just for pytest output
        return self.__str__()

    def get_document(self):
        from .Document import Document

        n = self
        while(n is not None and not isinstance(n, Document)):
            n = n._parent
        return n

    def get_type(self):
        '''
        Return the type of the node
        '''
        raise NotImplementedError('get_type has not been implemented in class ' + self.__class__.__name__)

    def escape(self, text):
        text = text.replace('&', '&amp;')
        text = text.replace('<', '&lt;')
        text = text.replace('>', '&gt;')
        return text

    def unescape(self, text):
        text = text.replace('&amp;', '&')
        text = text.replace('&lt;', '<')
        text = text.replace('&gt;', '>')
        text = text.replace('"', '&quot;')
        text = text.replace("'", '&apos;')
        return text

    def attached(self):
        pass

    def detached(self):
        pass

    def find_by_id(self, id_):
        return None

    def get_node_count(self):
        return 1

    def get_document_order(self):
        if self._parent is None:
            raise UnattachedElementException('Element ' + str(self) + ' is not attached to a document')

        do = self._parent.get_document_order()
        try:
            do += len(self._parent.namespace_nodes)
        except AttributeError:
            pass
        try:
            do += len(self._parent.attribute_nodes)
        except AttributeError:
            pass

        for c in self._parent.children:
            if c == self:
                return do
            do += c.get_node_count()

        raise ValueError('Unable to find node ' + str(self) + ' in ' + str(self._parent) + ' children')



#!/usr/bin/env python

# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import re
from glob import glob
from os.path import dirname, join

from setuptools import find_packages, setup

with open(join(dirname(__file__), 'VERSION.txt'), 'r') as f:
    VERSION = f.read().strip()

with open(join(dirname(__file__), 'README.rst'), 'r') as f:
    README = f.read()

with open(join(dirname(__file__), 'CHANGELOG.rst'), 'r') as f:
    CHANGELOG = f.read()

with open(join(dirname(__file__), 'CLASSIFIERS.txt'), 'r') as f:
    CLASSIFIERS = [line.strip() for line in f]

with open(join(dirname(__file__), 'KEYWORDS.txt'), 'r') as f:
    KEYWORDS = ' '.join([line.strip() for line in f])

with open(join(dirname(__file__), 'requirements.txt'), 'r') as f:
    REQUIREMENTS = [line.strip() for line in f]

long_description = README + '\n' + CHANGELOG

setup(name='expatriate',
    version=VERSION,
    license='LGPL',
    description='Library wrapping expat for parsing and generating XML',
    long_description=long_description,
    author='Casey Jaymes',
    author_email='cjaymes@gmail.com',
    url='https://github.com/cjaymes/expatriate',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    install_requires=REQUIREMENTS,
    tests_require=[
        'pytest',
    ],
    zip_safe=True,
    include_package_data=True,
)



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import logging
import re

from ..decorators import *
from .DateTimeType import DateTimeType
from .SevenPropertyModel import SevenPropertyModel

logger = logging.getLogger(__name__)

class DateTimeStampType(DateTimeType):
    def parse_value(self, value):
        m = re.fullmatch(r'(-?\d\d\d\d)-(\d\d)-(\d\d)T(\d\d):(\d\d):(\d\d(\.\d+)?)((([-+])(\d\d):(\d\d))|Z)', value)
        if not m:
            raise ValueError('Unable to parse DateTimeStampType value')

        # fudge for python 3.5 compatibility w/match obj __getitem__
        m = list(m.groups())
        m.insert(0, None)

        return SevenPropertyModel(year=m[1], month=m[2], day=m[3], hour=m[4], minute=m[5], second=m[6], timezoneOffset=m[8])



# Copyright 2016 Casey Jaymes

# This file is part of Expatriate.
#
# Expatriate is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Expatriate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Expatriate.  If not, see <http://www.gnu.org/licenses/>.

import importlib
import logging
import re

import expatriate
from publishsubscribe import PublishingDict, PublishingList

from .exceptions import *
from .Mapper import Mapper

logger = logging.getLogger(__name__)

class ElementMapper(Mapper):
    '''
        **kwargs**

        namespace
            The xml namespace to match. It can also be * to match any namespace.
            If not specified, it defaults to the parent element's namespace.
        local_name
            Required. The local name of the xml element we're matching. It can
            also be * to match any local name.

        into
            The python attribute to store the value of the element into.
            Defaults to the local_name if not specified.
        list
            The python attribute to store the value of the element into (as a
            list).  Defaults to the local_name if not specified.
        dict
            The python attribute to store the value of the element into (as a
            dict). Defaults to the local_name if not specified.

        dict_key
            The attribute of the sub-element to use as the key of the dict. By
            default it is the *id* attribute.
        dict_value
            The attribute of the sub-element to use as the value of the dict. By
            default it is the value of the element.

        Unless One of *type* or *cls* must be specified. *defer_class_load* can be
        used to load the class upon access instead of passing the class:

        type
            The type of the expected value. Types are stored directly as data,
            no enclosed in a model class. Types usually restrict the domain
            values.
        cls
            The model class with which to load the element.

        min
            The minimum number of elements to be present. Can be numeric or None
            (the  default).
        max
            The maximum number of elements to be present. Can be numeric or None
            (the default).

        enum
            Enumeration to which the value of the element must belong.
        pattern
            Pattern which the value of the element must match.

        nillable
            If True, the element can be nil (from the xsi spec). False
            specifies that it cannot (the default).
    '''

    def __init__(self, **kwargs):
        if 'local_name' not in kwargs:
            raise DecoratorException('Attributes need at least local_name defined')

        super().__init__(**kwargs)

    def get_attr_name(self):
        from .Model import Model

        if self._kwargs['local_name'] == Model.ANY_LOCAL_NAME:
            if 'into' not in self._kwargs:
                return '_elements'
            else:
                return self._kwargs['into']
        elif 'list' in self._kwargs:
            return self._kwargs['list']
        elif 'dict' in self._kwargs:
            return self._kwargs['dict']
        else:
            if 'into' in self._kwargs:
                return self._kwargs['into']
            else:
                return self._kwargs['local_name'].replace('-', '_')

    def initialize(self, model):
        from .Model import Model

        name = self.get_attr_name()

        if self._kwargs['local_name'] == Model.ANY_LOCAL_NAME:
            value = PublishingList()
            value.subscribe(model)
        elif 'list' in self._kwargs:
            value = PublishingList()
            value.subscribe(model)
        elif 'dict' in self._kwargs:
            value = PublishingDict()
            value.subscribe(model)
        else:
            value = None

        # initialze the attr if it doesn't exist
        if not hasattr(model, name):
            logger.debug('Initializing ' + str(self) + ' '+ name + ' to ' + str(value))
            setattr(model, name, value)

    def get_namespace(self):
        if 'namespace' in self._kwargs:
            return self._kwargs['namespace']
        else:
            return None

    def get_local_name(self):
        return self._kwargs['local_name']

    def find_reference_in(self, ref, model):
        from .Model import Model

        name = self.get_attr_name()
        attr = getattr(model, name)

        logger.debug('Looking for reference ' + ref + ' in ' + name + ' of ' + str(model))

        if isinstance(attr, list):
            for child in attr:
                try:
                    return child.find_reference(ref)
                except ReferenceException:
                    pass
        elif isinstance(attr, dict):
            for child in attr.values():
                try:
                    return child.find_reference(ref)
                except ReferenceException:
                    pass
        elif isinstance(attr, Model):
            return attr.find_reference(ref)

        raise ReferenceException('Could not find reference ' + ref
                + ' within ' + str(model) + '.' + name)

    def matches(self, el, model):
        from .Model import Model

        namespace = self.get_namespace()
        # if namespace is None:
        #     namespace = model.package_to_namespace(model.get_package())

        matches = (namespace, self.get_local_name()) in (
            (el.namespace, el.local_name),
            (None, el.local_name),
            (None, Model.ANY_LOCAL_NAME),
            (el.namespace, Model.ANY_LOCAL_NAME),
            (Model.ANY_NAMESPACE, Model.ANY_LOCAL_NAME)
        )

        if matches:
            logger.debug(str(self) + str((namespace, self.get_local_name())) + ' matches ' + str(el))
        else:
            logger.debug(str(self) + str((namespace, self.get_local_name())) + ' does not match ' + str(el))

        return matches

    def class_for_element(self, el, model):
        from .Model import Model

        if el.namespace is None:
            model_package = model.get_package()
        else:
            model_package = Model.namespace_to_package(el.namespace)

        if self.get_local_name() == Model.ANY_LOCAL_NAME:
            return Model.class_for_element(el)
        elif 'cls' in self._kwargs:
            return self._load_cls(self._kwargs['cls'])
        else:
            raise NotImplementedError

    def parse_in(self, model, el):
        from .Model import Model

        logger.debug('Parsing element ' + el.name + ' using kwargs: ' + str(self._kwargs))

        name = self.get_attr_name()

        if 'ignore' in self._kwargs and self._kwargs['ignore'] == True:
            logger.debug('Ignoring ' + name + ' element in ' + str(model))
            return

        if self._kwargs['local_name'] == Model.ANY_LOCAL_NAME:
            logger.debug(str(model) + ' parsing elements matching * into ' + name)

            lst = getattr(model, name)

            if el.is_nil():
                # check we can accept nil
                if 'nillable' not in self._kwargs or not self._kwargs['nillable']:
                    raise ValueError(str(el) + ' is nil, but not expecting nil value')
                value = None
            elif 'type' in self._kwargs:
                type_ = self._kwargs['type']()
                value = type_.parse_value(el.get_string_value())
            else:
                value = Model.load(model, el)

            lst.append(value)

            logger.debug('Appended ' + str(value) + ' to ' + name)

        elif 'list' in self._kwargs:
            logger.debug(str(model) + ' parsing ' + str(el) + ' elements into ' + name)

            lst = getattr(model, name)

            if el.is_nil():
                # check we can accept nil
                if 'nillable' not in self._kwargs or not self._kwargs['nillable']:
                    raise ValueError(str(el) + ' is nil, but not expecting nil value')
                value = None
            elif 'type' in self._kwargs:
                type_ = self._kwargs['type']
                value = el.get_string_value()
                value = type_().parse_value(value)
            else:
                value = Model.load(model, el)

            lst.append(value)

            logger.debug('Appended ' + str(value) + ' to ' + name)

        elif 'dict' in self._kwargs:
            logger.debug(str(model) + ' parsing ' + str(el) + ' elements into ' + name)

            dict_ = getattr(model, name)

            # TODO: implement key_element as well
            if 'dict_key' in self._kwargs:
                if self._kwargs['dict_key'] not in el.attributes:
                    key = None
                else:
                    key = el.attributes[self._kwargs['dict_key']]
            else:
                if 'id' not in el.attributes:
                    key = None
                else:
                    key = el.attributes['id']

            # TODO: implement value_element? as well
            if el.is_nil():
                # check we can accept nil
                if 'nillable' not in self._kwargs or not self._kwargs['nillable']:
                    raise ValueError(str(el) + ' is nil, but not expecting nil value')
                value = None
            elif 'dict_value' in self._kwargs:
                # TODO add ContentMapper
                # try parsing from an attribute
                if self._kwargs['dict_value'] not in el.attributes:
                    raise ValueError('Could not parse value from '
                        + str(el) + ' attribute '
                        + self._kwargs['dict_value'])

                if 'type' not in self._kwargs:
                    raise ValueError('Could not parse value from '
                        + str(el) + ' attribute '
                        + self._kwargs['dict_value']
                        + ' without explicit type')

                type_ = self._kwargs['type']()
                value = el.attributes[self._kwargs['dict_value']]
                value = type_.parse_value(value)

            else:
                # try parsing from the element itself, just mapping with the key
                if 'type' in self._kwargs:
                    type_ = self._kwargs['type']()
                    value = type_.parse_value(el.get_string_value())
                else:
                    # needs 'cls' in self._kwargs
                    value = Model.load(model, el)

            dict_[key] = value

            logger.debug('Mapped ' + str(key) + ' to ' + str(value) + ' in ' + name)

        elif 'cls' in self._kwargs:
            logger.debug(str(model) + ' parsing ' + str(el) + ' element as '
                + str(self._kwargs['cls']))

            if el.is_nil():
                # check we can accept nil
                if 'nillable' not in self._kwargs or not self._kwargs['nillable']:
                    raise ValueError(str(el) + ' is nil, but not expecting nil value')
                value = None
            else:
                value = Model.load(model, el)

            setattr(model, name, value)

            logger.debug('Set attribute ' + str(name) + ' to ' + str(value)
                + ' in ' + str(model))

        elif 'type' in self._kwargs:
            logger.debug(str(model) + ' parsing ' + str(el) + ' elements as '
                + self._kwargs['type'])

            if el.is_nil():
                # check we can accept nil
                if 'nillable' not in self._kwargs or not self._kwargs['nillable']:
                    raise ValueError(str(el) + ' is nil, but not expecting nil value')
                value = None
            else:
                type_ = self._kwargs['type']
                # load from a tuple (module_name, class_name) to defer class load
                if isinstance(type_, tuple):
                    mod = importlib.import_module(type_[0])
                    type_ = getattr(mod, type_[1])
                value = el.get_string_value()
                value = type_().parse_value(value)

            setattr(model, name, value)

            logger.debug('Set attribute ' + str(name) + ' to ' + str(value)
                + ' in ' + str(model))

        elif 'enum' in self._kwargs:
            logger.debug(str(model) + ' parsing ' + str(el)
                + ' elements from enum ' + str(self._kwargs['enum']))

            value = el.get_string_value()
            if value not in self._kwargs['enum']:
                raise EnumerationException(str(el)
                    + ' value must be one of ' + str(self._kwargs['enum']))

            setattr(model, name, value)

            logger.debug('Set enum attribute ' + str(name) + ' to '
                + str(value) + ' in ' + str(model))

        elif 'pattern' in self._kwargs:
            logger.debug(str(model) + ' parsing ' + str(el)
                + ' elements from pattern ' + str(self._kwargs['pattern']))

            value = el.get_string_value()
            if re.match(self._kwargs['pattern'], value) is None:
                raise PatternException(str(el)
                    + ' value must match ' + str(self._kwargs['pattern']))

            setattr(model, name, value)

            logger.debug('Set pattern attribute ' + str(name) + ' to '
                + str(value) + ' in ' + str(model))

        else:
            raise UnknownElementException(str(model) + ' could not parse '
                + str(el) + ' element')

    def validate(self, model):
        from .Model import Model

        name = self.get_attr_name()

        min_ = 1
        if (
            'dict' in self._kwargs
            or 'list' in self._kwargs
            or self.get_local_name() == Model.ANY_LOCAL_NAME
        ):
            count = len(getattr(model, name))
            # dicts and lists default to no max
            max_ = None
        else:
            max_ = 1
            count = 1

        # if there's an explicit min/max definition, use that
        if 'min' in self._kwargs:
            min_ = self._kwargs['min']
        if 'max' in self._kwargs:
            max_ = self._kwargs['max']

        # check that we have the min & max of those elements
        if min_ != 0 and count < min_:
            raise MinimumElementException(str(model)
                + ' must have at least ' + str(min_) + ' '
                + str((self.get_namespace(), self.get_local_name()))
                + ' elements')

        if max_ is not None and count > max_:
            raise MaximumElementException(str(model)
                + ' may have at most ' + str(max_) + ' '
                + str((self.get_namespace(), self.get_local_name()))
                + ' elements')

    def setattr(self, model, name, value):
        from .Model import Model

        if not hasattr(model, '_initialized'):
            object.__setattr__(model, name, value)
            return

        if (
            'list' in self._kwargs
            or self.get_local_name() == Model.ANY_LOCAL_NAME
        ):
            # wrap in PublishingList; updates are captured via Publisher
            object.__setattr__(model, name, PublishingList(value))
        elif 'dict' in self._kwargs:
            # wrap in PublishingDict; updates are captured via Publisher
            object.__setattr__(model, name, PublishingDict(value))
        else:
            old_value = getattr(model, name)
            object.__setattr__(model, name, value)
            if old_value is None and value is None:
                pass
            elif old_value is None and value is not None:
                model.data_added(value, None, value)
            elif old_value is not None and value is None:
                model.data_deleted(value, None, value)
            else: # old_value is not None and value is not None
                model.data_updated(value, None, old_value, value)

    def produce_in(self, el, model, id_):
        from .Model import Model

        name = self.get_attr_name()
        attr = getattr(model, name)
        if id_ is None:
            value = attr
        else:
            value = attr[id_]

        logger.debug(str(self) + ' producing ' + str(model) + ' element '
            + name  + str([id_]) + ' according to ' + str(self._kwargs))

        if hasattr(value, '_namespace') and value._namespace is not None:
            namespace = value._namespace
        elif 'namespace' in self._kwargs:
            if self._kwargs['namespace'] == Model.ANY_NAMESPACE:
                raise ElementMappingException(str(self) + ' cannot map '
                    + str(model) + '.' + name + ' without a namespace override')
            namespace = self._kwargs['namespace']
        else:
            namespace = el.namespace

        if hasattr(value, '_prefix') and value._prefix is not None:
            prefix = value._prefix
        else:
            prefix = el.namespace_to_prefix(namespace)

        if hasattr(value, '_local_name') and value._local_name is not None:
            local_name = value._local_name
        else:
            if self._kwargs['local_name'] == Model.ANY_LOCAL_NAME:
                raise ElementMappingException(str(self) + ' cannot map '
                    + str(model) + '.' + name + ' without a local_name override')
            local_name = self._kwargs['local_name']

        if local_name is None:
            raise ElementMappingException('local_name must be defined by constructor or @element')

        if self._kwargs['local_name'] == Model.ANY_LOCAL_NAME:
            if value is None:
                return

            el.append(value.produce(parent_el=el))
        elif 'list' in self._kwargs:
            if 'type' in self._kwargs:
                if value is None:
                    sub_el = expatriate.Element(local_name, namespace=namespace, prefix=prefix, parent=el)
                    sub_el.attributes['xmlns:xsi'] = 'http://www.w3.org/2001/XMLSchema-instance'
                    sub_el.attributes['xsi:nil'] = 'true'
                    el.append(sub_el)
                else:
                    # wrap value in xs element
                    class_ = self._kwargs['type']
                    child = class_(local_name=local_name, namespace=namespace, prefix=prefix)
                    child.set_value(value)
                    el.append(child.produce(parent_el=el))
            elif 'cls' in self._kwargs:
                if value is None:
                    sub_el = expatriate.Element(local_name, namespace=namespace, prefix=prefix, parent=el)
                    sub_el.attributes['xmlns:xsi'] = 'http://www.w3.org/2001/XMLSchema-instance'
                    sub_el.attributes['xsi:nil'] = 'true'
                    el.append(sub_el)
                elif not isinstance(value, self._load_cls(self._kwargs['cls'])):
                    raise ElementMappingException('Value ' + str(value)
                        + ' is not of the expected class: '
                        + str(self._kwargs['cls']))
                else:
                    el.append(value.produce(parent_el=el))

            else:
                raise ElementMappingException('"cls" or "type" must be '
                    + 'defined for "list" model mapping')

        elif 'dict' in self._kwargs:
            # TODO: implement key_element as well
            if 'dict_key' in self._kwargs:
                key_name = self._kwargs['dict_key']
            else:
                key_name = 'id'

            if 'type' in self._kwargs:
                sub_el = expatriate.Element(local_name, namespace=namespace, prefix=prefix, parent=el)
                sub_el.attributes[key_name] = id_
                if 'dict_value' in self._kwargs:
                    if value is None:
                        raise ValueError(str(self)
                            + ' Cannot have none for a dict_value: '
                            + self._kwargs['dict'] + '[' + id_ + ']')
                    type_ = self._kwargs['type']()
                    value = type_.produce_value(value)
                    sub_el.attributes[self._kwargs['dict_value']] = value
                else:
                    if value is None:
                        sub_el.attributes['xmlns:xsi'] = 'http://www.w3.org/2001/XMLSchema-instance'
                        sub_el.attributes['xsi:nil'] = 'true'
                    else:
                        type_ = self._kwargs['type']()
                        sub_el.children.append(expatriate.CharacterData(type_.produce_value(value)))
                el.append(sub_el)

            elif 'cls' in self._kwargs:
                if value is None:
                    sub_el = expatriate.Element(local_name, namespace=namespace, prefix=prefix, parent=el)
                    sub_el.attributes[key_name] = id_
                    sub_el.attributes['xmlns:xsi'] = 'http://www.w3.org/2001/XMLSchema-instance'
                    sub_el.attributes['xsi:nil'] = 'true'
                    el.append(sub_el)
                else:
                    setattr(value, key_name, id_)
                    el.append(value.produce(parent_el=el))

            else:
                raise ValueError('"class" or "type" must be defined for "dict" model mapping')

        elif 'cls' in self._kwargs:
            if value is None:
                return

            el.append(value.produce(parent_el=el))

        elif 'type' in self._kwargs:
            if value is None:
                return

            class_ = self._kwargs['type']
            child = class_(local_name=local_name, namespace=namespace, prefix=prefix)
            child.set_value(value)

            el.append(child.produce(parent_el=el))

        elif 'enum' in self._kwargs:
            if value is None:
                return

            if value not in self._kwargs['enum']:
                raise EnumerationException(str(namespace, local_name)
                    + ' value must be one of ' + str(self._kwargs['enum'])
                    + ': ' + str(value))

            child = StringType(local_name=local_name, namespace=namespace, prefix=prefix)
            child.set_value(value)

            el.append(child.produce(parent_el=el))

        else:
            raise UnknownElementException(str(self) + ' could not produce ' + str(namespace, local_name) + ' element')
